/*
 * swo_log.h
 *
 * Created on: 20.01.2025
 * Author: Kilian
 *
 * Use the SWO interface to log messages
 */
#pragma once
#ifndef __SWO_LOG_H
#define __SWO_LOG_H
#include "stm32h7xx_hal.h"
#include <stdarg.h>
#include <stdio.h>

#define MAX_MESSAGE_LENGTH 256

enum log_level_t {
  LOG_LEVEL_NOISY,
  LOG_LEVEL_DEBUG,
  LOG_LEVEL_INFO,
  LOG_LEVEL_WARNING,
  LOG_LEVEL_ERROR,
  LOG_LEVEL_FATAL
};

[[maybe_unused]] static const char * const log_level_names[] = {
    "\033[37m[NOISY]\033[0m ",
    "\033[37m[DEBUG]\033[0m ",
    "\033[97m[INFO] \033[0m ",
    "\033[93m[WARN] \033[0m ",
    "\033[31m[ERROR]\033[0m ",
    "\033[31m[FATAL]\033[0m "
};

static inline bool __ITM_channel_enabled(uint32_t channel) {
  return (ITM->TER & (1UL << channel)) != 0UL;
}

// adapted from ITM_SendChar() in the CMSIS-Core
//
// channel should be between 0 and 31
static inline uint32_t __swo_putc(uint32_t c, unsigned int channel) {
  if (((ITM->TCR & ITM_TCR_ITMENA_Msk) != 0UL) && /* ITM enabled */
      ((ITM->TER & (1UL << channel)) != 0UL))     /* ITM Port enabled */
  {
    while (ITM->PORT[channel].u32 == 0UL) {
      __NOP();
    }
    ITM->PORT[channel].u8 = (uint8_t)c;
  }
  return (c);
}

#define DEBUG_CHANNEL_ENABLED(channel)                                         \
  ({                                                                           \
    unsigned int ch = (channel);                                               \
    (ch < 32) ? __ITM_channel_enabled(ch) : false;                             \
  })

[[gnu::nonnull(2), gnu::null_terminated_string_arg(2)]]
static inline void __swo_print(unsigned int channel, const char *str) {
  if (!__ITM_channel_enabled(channel)) {
    return;
  }
  while (*str) {
    __swo_putc(*str++, channel);
  }
}

#define debug_log(level, msg, ...)                                             \
  do {                                                                         \
    if (DEBUG_CHANNEL_ENABLED(level)) {                                        \
      char buffer[MAX_MESSAGE_LENGTH];                                         \
      size_t len = snprintf(buffer, sizeof(buffer), msg, ##__VA_ARGS__);       \
      __swo_putc('\n', level);                                                 \
      __swo_print(level, log_level_names[level]);                              \
      __swo_print(level, buffer);                                              \
      if (len >= sizeof(buffer)) {                                             \
        __swo_print(level, " [message length exceeded] ");                     \
      }                                                                        \
    }                                                                          \
  } while (0)

#define debug_log_cont(level, msg, ...)                                        \
  do {                                                                         \
    if (DEBUG_CHANNEL_ENABLED(level)) {                                        \
      char buffer[MAX_MESSAGE_LENGTH];                                         \
      size_t len = snprintf(buffer, sizeof(buffer), msg, ##__VA_ARGS__);       \
      __swo_print(level, buffer);                                              \
      if (len >= sizeof(buffer)) {                                             \
        __swo_print(level, " [message length exceeded] ");                     \
      }                                                                        \
      __swo_putc('\n', level);                                                 \
    }                                                                          \
  } while (0)

#endif /* __SWO_LOG_H */
